import logging
from collections import defaultdict

from django.db import transaction
from django.db.models.manager import Manager
from rest_framework import serializers
from rest_framework.utils import html

from core import models
from core.models import Feedback
from util.factories import GradyUserFactory

from .generic import DynamicFieldsModelSerializer

log = logging.getLogger(__name__)
user_factory = GradyUserFactory()


class FeedbackCommentDictionarySerializer(serializers.ListSerializer):

    def to_internal_value(self, comment_dict):
        """ Converts a line_no -> comment list dictionary back to a list
        of comments. Currently we do not have any information about the
        feedback since it is not availiable in this scope. Feedback is
        responsible to add it later on update/creation """
        if html.is_html_input(comment_dict):
            comment_dict = html.parse_html_list(comment_dict)

        if not isinstance(comment_dict, dict):
            raise serializers.ValidationError(
                'Comments have to be provided as a dict'
                'with: line -> list of comments'
            )

        ret = []
        errors = []

        for line, comment in comment_dict.items():
            try:
                comment['of_line'] = line
                validated = self.child.run_validation(comment)
            except serializers.ValidationError as err:
                errors.append(err.detail)
            else:
                ret.append(validated)
                errors.append({})

        if any(errors):
            raise serializers.ValidationError(errors)

        return ret

    def to_representation(self, comments):
        """ Provides a dict where all the keys correspond to lines and contain
        a list of comments on that line. """
        if isinstance(comments, Manager):
            comments = comments.all()

        ret = defaultdict(list)
        for comment in comments:
            ret[comment.of_line].append(self.child.to_representation(comment))
        return ret


class FeedbackCommentSerializer(DynamicFieldsModelSerializer):
    of_tutor = serializers.StringRelatedField(source='of_tutor.username')

    class Meta:
        model = models.FeedbackComment
        fields = ('pk',
                  'text',
                  'created',
                  'of_tutor',
                  'of_line',
                  'visible_to_student')
        read_only_fields = ('created', 'of_tutor')
        extra_kwargs = {
            'of_feedback': {'write_only': True},
            'of_line': {'write_only': True},
        }
        list_serializer_class = FeedbackCommentDictionarySerializer


class FeedbackSerializer(DynamicFieldsModelSerializer):
    feedback_lines = FeedbackCommentSerializer(many=True, required=False)
    of_submission_type = serializers.ReadOnlyField(
        source='of_submission.type.pk')
    feedback_stage_for_user = serializers.SerializerMethodField()

    def get_feedback_stage_for_user(self, obj):
        """ Search for the assignment of this feedback and report in which
        stage the tutor has worked on it.

        Note: This method is unorthodox since it mingles the rather dump
        feedback object with assignment logic. The reverse lookups in the
        method are not pre-fetched. Remove if possible. """
        if 'request' not in self.context:
            return

        # This is only required for tutors
        user = self.context['request'].user
        if user.role == models.UserAccount.REVIEWER:
            return None

        assignments = obj.of_submission.assignments.filter(
            subscription__owner=user)

        if assignments.count() == 0:
            return None

        return assignments[0].subscription.feedback_stage

    @transaction.atomic
    def create(self, validated_data) -> Feedback:
        submission = validated_data.pop('of_submission')
        feedback_lines = validated_data.pop('feedback_lines', [])

        feedback = Feedback.objects.create(of_submission=submission,
                                           **validated_data)
        submission.meta.feedback_authors.add(self.context['request'].user)

        for comment in feedback_lines:
            models.FeedbackComment.objects.create(
                of_feedback=feedback,
                of_tutor=self.context['request'].user,
                **comment
            )

        return Feedback.objects.get(of_submission=submission)

    @transaction.atomic
    def update(self, feedback, validated_data):
        for comment in validated_data.pop('feedback_lines', []):
            models.FeedbackComment.objects.update_or_create(
                of_feedback=feedback,
                of_tutor=self.context['request'].user,
                of_line=comment.get('of_line'),
                defaults={'text': comment.get('text')})

        return super().update(feedback, validated_data)

    def validate_of_submission(self, submission):
        feedback = self.instance
        if feedback is not None and feedback.of_submission.pk != submission.pk:
            raise serializers.ValidationError(
                'It is not allowed to update this field.')

        return submission

    def validate(self, data):
        if self.instance:
            score = data.get('score', self.instance.score)
            submission = data.get('of_submission', self.instance.of_submission)
        else:
            try:
                score = data.get('score')
                submission = data.get('of_submission')
            except KeyError:
                raise serializers.ValidationError(
                    'You need a score and a submission.')

        if not 0 <= score <= submission.type.full_score:
            raise serializers.ValidationError(
                f'Score has to be in range [0..{submission.type.full_score}].')

        has_full_score = score == submission.type.full_score
        has_feedback_lines = ('feedback_lines' in data and
                              len(data['feedback_lines']) > 0 or
                              self.instance is not None and
                              self.instance.feedback_lines.count() > 0)

        if not has_full_score and not has_feedback_lines:
            raise serializers.ValidationError(
                'Sorry, you have to explain why this does not get full score')

        http_method = self.context['request'].method
        if hasattr(submission, 'feedback') and http_method == 'POST':
            raise serializers.ValidationError(
                'Feedback for this submission already exists')

        for comment in data.get('feedback_lines', {}):
            lines_in_submission = len(submission.text.split('\n'))
            if not 0 < comment['of_line'] <= lines_in_submission:
                raise serializers.ValidationError(
                    "Cannot comment line number %d of %d" % (
                        comment['of_line'], lines_in_submission))

        return data

    class Meta:
        model = Feedback
        fields = ('pk', 'of_submission', 'is_final', 'score', 'feedback_lines',
                  'created', 'of_submission_type', 'feedback_stage_for_user')


class VisibleCommentFeedbackSerializer(FeedbackSerializer):
    feedback_lines = serializers.SerializerMethodField()
    of_submission_type = serializers.ReadOnlyField(
        source='of_submission.type.pk')

    def get_feedback_lines(self, feedback):
        comments = feedback.feedback_lines.filter(visible_to_student=True)
        serializer = FeedbackCommentSerializer(
            comments,
            many=True,
            fields=('pk', 'text', 'created', 'of_line',)
        )
        # this is a weird hack because, for some reason, serializer.data
        # just won't contain the correct data. Instead .data returns a list
        # containing just the `of_line` attr of the serialized comments
        # after long debugging i found that for inexplicable reasons
        # `data.serializer._data` contains the correct data. No clue why.
        return serializer.data.serializer._data

    class Meta:
        model = Feedback
        fields = ('pk', 'of_submission', 'is_final', 'score', 'feedback_lines',
                  'created', 'of_submission_type')
