import unittest

from rest_framework import status
from rest_framework.test import APIRequestFactory, APITestCase

from core import models
from core.models import Feedback, FeedbackComment, Submission, SubmissionType
from util.factories import GradyUserFactory, make_test_data


class FeedbackRetrieveTestCase(APITestCase):

    factory = GradyUserFactory()
    EXPECTED_SCORE = 23

    @classmethod
    def setUpTestData(cls):
        cls.score = 23
        cls.tutor = cls.factory.make_tutor()
        cls.student = cls.factory.make_student()
        cls.reviewer = cls.factory.make_reviewer()
        cls.tutors = [cls.tutor, cls.reviewer]
        cls.request_factory = APIRequestFactory()
        cls.submission_type = SubmissionType.objects.create(
            name='Cooking some crystal with Jesse')
        cls.sub = Submission.objects.create(student=cls.student.student,
                                            type=cls.submission_type)
        cls.feedback = Feedback.objects.create(score=23, is_final=False,
                                               of_submission=cls.sub)

        for line in range(1, 3):
            for tutor in cls.tutors:
                FeedbackComment.objects.create(text='fortytwo',
                                               of_feedback=cls.feedback,
                                               of_tutor=tutor,
                                               of_line=line)

    def setUp(self):
        self.client.force_authenticate(user=self.reviewer)
        self.response = self.client.get(f'/api/feedback/{self.sub.pk}/')
        self.data = self.response.data

    def test_only_one_final_comment_per_line(self):
        comments_on_first_line = FeedbackComment.objects.filter(of_line=1)
        self.assertEqual(2, comments_on_first_line.count())
        final_comments = [comment for comment in comments_on_first_line.all()
                          if comment.visible_to_student]
        self.assertEqual(1, len(final_comments))

    def test_can_retrieve_feedback_via_endpoint(self):
        self.assertEqual(self.response.status_code, status.HTTP_200_OK)

    def test_if_feedback_contains_correct_score(self):
        self.assertIn('score', self.data)
        self.assertEqual(self.data.get('score'), self.EXPECTED_SCORE)

    def test_if_feedback_contains_linekeys(self):
        self.assertIn('feedback_lines', self.data)
        self.assertIn(1, self.data['feedback_lines'])
        self.assertIn(2, self.data['feedback_lines'])

    def test_if_feedback_contains_final(self):
        self.assertIn('is_final', self.data)
        self.assertIsNotNone(self.data['is_final'])

    def test_if_comment_contains_text(self):
        self.assertIn('text', self.data['feedback_lines'][1][0])
        self.assertEqual(
            'fortytwo', self.data['feedback_lines'][1][0]['text'])

    def test_if_comment_contains_created(self):
        self.assertIn('created', self.data['feedback_lines'][1][0])
        self.assertIsNotNone(self.data['feedback_lines'][1][0]['created'])

    def test_if_comment_has_tutor(self):
        self.assertIn('of_tutor', self.data['feedback_lines'][1][0])
        self.assertEqual(
            self.tutor.username,
            self.data['feedback_lines'][1][0]['of_tutor'])

    def test_if_comment_has_final(self):
        self.assertIn('visible_to_student', self.data['feedback_lines'][1][0])
        self.assertIsNotNone(
            self.data['feedback_lines'][1][0]['visible_to_student'])


class FeedbackCreateTestCase(APITestCase):

    @classmethod
    def setUpTestData(cls):
        cls.url = '/api/feedback/'
        cls.user_factory = GradyUserFactory()
        cls.tutor = cls.user_factory.make_tutor(password='p')
        cls.student = cls.user_factory.make_student()
        cls.submission_type = SubmissionType.objects.create(
            name='Cooking some crystal with Jesse',
            full_score=100
        )
        text = ''' First line of defense
        We do not have a second line
        security via obscurity
        is very bad. '''
        cls.sub = Submission.objects.create(student=cls.student.student,
                                            type=cls.submission_type,
                                            text=text)

    def setUp(self):
        self.client.force_authenticate(user=self.tutor)
        self.subscription = models.SubmissionSubscription.objects.create(
            owner=self.tutor,
            query_type='random'
        )
        self.assignment = self.subscription.get_or_create_work_assignment()

    def test_cannot_create_feedback_without_feedback_lines(self):
        # TODO this test has to be adapted to test the various constraints
        # e.g. feedback without lines can only be given if the score is equal
        # to the max Score for this submission
        data = {
            'score': 10,
            'is_final': False,
            'of_submission': self.assignment.submission.pk

        }
        self.assertEqual(Feedback.objects.count(), 0)
        response = self.client.post(self.url, data, format='json')
        self.assertEqual(status.HTTP_400_BAD_REQUEST, response.status_code)
        self.assertEqual(Feedback.objects.count(), 0)

    def test_cannot_create_feedback_with_score_higher_than_max(self):
        data = {
            'score': 101,
            'is_final': False,
            'of_submission': self.assignment.submission.pk
        }
        self.assertEqual(Feedback.objects.count(), 0)
        response = self.client.post(self.url, data, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(Feedback.objects.count(), 0)

    def test_tutor_cannot_set_feedback_final_on_creation(self):
        data = {
            'score': 100,
            'is_final': True,
            'of_submission': self.assignment.submission.pk
        }
        response = self.client.post(self.url, data, format='json')
        self.assertEqual(status.HTTP_403_FORBIDDEN, response.status_code)
        self.assertEqual(Feedback.objects.count(), 0)

    def test_tutor_has_to_write_a_line_if_score_is_not_100_percent(self):
        data = {
            'score': 50,
            'is_final': False,
            'of_submission': self.assignment.submission.pk
        }
        response = self.client.post(self.url, data, format='json')
        self.assertEqual(status.HTTP_400_BAD_REQUEST, response.status_code)
        self.assertEqual(Feedback.objects.count(), 0)

    def test_cannot_create_feedback_with_score_less_than_zero(self):
        data = {
            'score': -1,
            'is_final': False,
            'of_submission': self.assignment.submission.pk
        }
        self.assertEqual(Feedback.objects.count(), 0)
        response = self.client.post(self.url, data, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(Feedback.objects.count(), 0)

    def test_check_score_is_set_accordingly(self):
        data = {
            'score': 5,
            'is_final': False,
            'of_submission': self.assignment.submission.pk,
            'feedback_lines': {
                '4': {
                    'text': 'Why you no learn how to code, man?'
                }
            }
        }
        self.client.post(self.url, data, format='json')
        object_score = self.sub.feedback.score
        self.assertEqual(object_score, 5)

    def test_can_create_feedback_with_comment(self):
        data = {
            'score': 0,
            'is_final': False,
            'of_submission': self.assignment.submission.pk,
            'feedback_lines': {
                '4': {
                    'text': 'Nice meth!'
                }
            }
        }
        self.assertEqual(FeedbackComment.objects.count(), 0)
        response = self.client.post(self.url, data, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(FeedbackComment.objects.count(), 1)

    def test_feedback_comment_is_created_correctly(self):
        data = {
            'score': 0,
            'is_final': False,
            'of_submission': self.assignment.submission.pk,
            'feedback_lines': {
                '3': {
                    'text': 'Nice meth!'
                }
            }
        }
        self.client.post(self.url, data, format='json')
        comment = FeedbackComment.objects.first()
        self.assertEqual(comment.of_tutor, self.tutor)
        self.assertEqual(comment.text, 'Nice meth!')
        self.assertIsNotNone(comment.created)
        self.assertEqual(comment.of_line, 3)
        self.assertTrue(comment.visible_to_student)

    def test_cannot_create_without_assignment(self):
        data = {
            'score': 0,
            'of_submission': self.assignment.submission.pk,
            'feedback_lines': {
                '2': {
                    'text': 'Well, at least you tried.'
                },
            }
        }
        self.assignment.delete()
        response = self.client.post(self.url, data, format='json')
        self.assertEqual(status.HTTP_403_FORBIDDEN, response.status_code)

    def test_cannot_create_with_someoneelses_assignment(self):
        data = {
            'score': 0,
            'of_submission': self.assignment.submission.pk,
            'feedback_lines': {
                '1': {
                    'text': 'Well, at least you tried.'
                },
            }
        }
        other_tutor = self.user_factory.make_tutor('Berta')
        self.client.force_authenticate(other_tutor)
        response = self.client.post(self.url, data, format='json')
        self.assertEqual(status.HTTP_403_FORBIDDEN, response.status_code)

    def test_can_create_multiple_feedback_comments(self):
        data = {
            'score': 0,
            'is_final': False,
            'of_submission': self.assignment.submission.pk,
            'feedback_lines': {
                '1': {
                    'text': 'Nice meth!'
                },
                '3': {
                    'text': 'Good one!'
                }
            }
        }
        self.client.post(self.url, data, format='json')
        first_comment = FeedbackComment.objects.get(text='Nice meth!')
        self.assertEqual(first_comment.of_tutor, self.tutor)
        self.assertIsNotNone(first_comment.created)
        self.assertEqual(first_comment.of_line, 1)
        self.assertTrue(first_comment.visible_to_student)

        second_comment = FeedbackComment.objects.get(text='Good one!')
        self.assertEqual(second_comment.of_tutor, self.tutor)
        self.assertIsNotNone(second_comment.created)
        self.assertEqual(second_comment.of_line, 3)
        self.assertTrue(second_comment.visible_to_student)


class FeedbackPatchTestCase(APITestCase):

    @classmethod
    def setUpTestData(cls):
        cls.burl = '/api/feedback/'
        cls.data = make_test_data({
            'submission_types': [
                {
                    'name': '01. Sort this or that',
                    'full_score': 35,
                    'description': 'Very complicated',
                    'solution': 'Trivial!'
                }],
            'students': [
                {'username': 'student01'}
            ],
            'tutors': [
                {'username': 'tutor01'},
                {'username': 'tutor02'}
            ],
            'reviewers': [{
                'username': 'reviewer01',
            }],
            'submissions': [{
                'text': 'function blabl\n'
                        '   on multi lines\n'
                        '       for blabla in bla:\n',
                'type': '01. Sort this or that',
                'user': 'student01'
            }]
        })

    def setUp(self):
        self.tutor01 = self.data['tutors'][0]
        self.tutor02 = self.data['tutors'][1]
        self.client.force_authenticate(user=self.tutor01)
        self.subscription = models.SubmissionSubscription.objects.create(
            owner=self.tutor01,
            query_type='random',
        )
        self.assignment = self.subscription.get_or_create_work_assignment()
        data = {
            'score': 35,
            'is_final': False,
            'of_submission': self.assignment.submission.pk,
            'feedback_lines': {
                '2': {'text': 'Very good.'},
            }
        }
        response = self.client.post(self.burl, data, format='json')
        self.feedback = Feedback.objects.get(
            of_submission=response.data['of_submission'])
        self.url = f'{self.burl}{self.feedback.of_submission.submission_id}/'

    def test_can_patch_onto_the_own_feedback(self):
        data = {
            'feedback_lines': {
                '1': {'text': 'Spam spam spam'},
            }
        }
        response = self.client.patch(self.url, data, format='json')
        self.assertEqual(status.HTTP_200_OK, response.status_code)
        self.assertEqual(
            'Spam spam spam',
            response.data['feedback_lines'][1][0]['text']
        )
        self.assertEqual(
            'Very good.',
            response.data['feedback_lines'][2][0]['text']
        )

    def test_can_update_a_single_line(self):
        data = {
            'feedback_lines': {
                '2': {'text': 'Turns out this is rather bad.'},
            }
        }

        response = self.client.patch(self.url, data, format='json')
        self.assertEqual(status.HTTP_200_OK, response.status_code)

    @unittest.expectedFailure
    def test_tutor_can_not_update_when_there_is_a_new_assignment(self):
        # Step 1 - Create a new assignment for Tutor 2
        second_subs = models.SubmissionSubscription.objects.create(
            owner=self.tutor02,
            query_type='random',
            feedback_stage='feedback-validation'
        )
        second_subs.get_or_create_work_assignment()

        # Step 2 - Tutor 1 tries to patch
        data = {
            'feedback_lines': {
                '2': {'text': 'Turns out this is rather bad.'},
            }
        }

        response = self.client.patch(self.url, data, format='json')
        self.assertEqual(status.HTTP_403_FORBIDDEN, response.status_code)

    def test_cannot_patch_first_feedback_final(self):
        data = {
            'feedback_lines': {
                '2': {'text': 'Turns out this is rather bad.'},
            },
            'is_final': True
        }

        response = self.client.patch(self.url, data, format='json')
        self.assertEqual(status.HTTP_403_FORBIDDEN, response.status_code)


class FeedbackCommentApiEndpointTest(APITestCase):

    @classmethod
    def setUpTestData(cls):
        cls.burl = '/api/feedback/'
        cls.data = make_test_data({
            'submission_types': [
                {
                    'name': '01. Sort this or that',
                    'full_score': 35,
                    'description': 'Very complicated',
                    'solution': 'Trivial!'
                }],
            'students': [
                {'username': 'student01'}
            ],
            'tutors': [
                {'username': 'tutor01'},
                {'username': 'tutor02'},
            ],
            'reviewers': [
                {'username': 'reviewer01'},
            ],
            'submissions': [{
                'text': 'function blabl\n'
                        '   on multi lines\n'
                        '       for blabla in bla:\n',
                'type': '01. Sort this or that',
                'user': 'student01',
                'feedback': {
                    'score': 5,
                    'is_final': True,
                    'feedback_lines': {
                        '1': [{'text': 'This is very bad!',
                               'of_tutor': 'tutor01'}],
                        '2': [{'text': 'And this is even worse!',
                               'of_tutor': 'tutor02'}],
                    }
                }
            }]
        })

    def setUp(self):
        self.url = '/api/feedback-comment/%s/'
        self.tutor01 = self.data['tutors'][0]
        self.tutor02 = self.data['tutors'][1]

    def test_tutor_can_delete_own_comment(self):
        self.client.force_authenticate(user=self.tutor01)
        comment = FeedbackComment.objects.get(of_tutor=self.tutor01)
        response = self.client.delete(self.url % comment.pk)
        self.assertEqual(status.HTTP_204_NO_CONTENT, response.status_code)

    def test_tutor_cannot_delete_foreign_comment(self):
        self.client.force_authenticate(user=self.tutor02)
        comment = FeedbackComment.objects.get(of_tutor=self.tutor02)
        self.client.force_authenticate(self.tutor01)
        response = self.client.delete(self.url % comment.pk)
        self.assertEqual(status.HTTP_404_NOT_FOUND, response.status_code)

    def test_reviewer_can_delete_everything_they_want(self):
        reviewer = self.data['reviewers'][0]
        self.client.force_authenticate(user=reviewer)
        comment01 = FeedbackComment.objects.get(of_tutor=self.tutor01)
        comment02 = FeedbackComment.objects.get(of_tutor=self.tutor02)

        response = self.client.delete(self.url % comment01.pk)
        self.assertEqual(status.HTTP_204_NO_CONTENT, response.status_code)

        response = self.client.delete(self.url % comment02.pk)
        self.assertEqual(status.HTTP_204_NO_CONTENT, response.status_code)
        try:
            FeedbackComment.objects.get(of_tutor=self.tutor01)
            FeedbackComment.objects.get(of_tutor=self.tutor02)
        except FeedbackComment.DoesNotExist:
            pass
        else:
            self.fail('No exception raised')

    def test_reviewer_can_set_comment_visibility(self):
        reviewer = self.data['reviewers'][0]
        self.client.force_authenticate(user=reviewer)
        comment = FeedbackComment.objects.get(of_tutor=self.tutor01)
        self.assertTrue(comment.visible_to_student)

        data = {
            'visible_to_student': False
        }

        response = self.client.patch(self.url % comment.pk, data)

        self.assertFalse(response.data['visible_to_student'])
        comment.refresh_from_db()
        self.assertFalse(comment.visible_to_student)
