from rest_framework import status
from rest_framework.test import APIClient, APITestCase

from core import models
from core.models import (Submission, SubmissionSubscription, SubmissionType,
                         SubscriptionEnded, SubscriptionTemporarilyEnded, TutorSubmissionAssignment)
from util.factories import GradyUserFactory, make_test_data


class SubmissionSubscriptionRandomTest(APITestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user_factory = GradyUserFactory()

    def setUp(self):
        self.t = self.user_factory.make_tutor()
        self.s1 = self.user_factory.make_student()
        self.s2 = self.user_factory.make_student()

        self.submission_type = SubmissionType.objects.create(
            name='submission_01', full_score=14)
        self.submission_01 = Submission.objects.create(
            type=self.submission_type, student=self.s1.student,
            text='I really failed')
        self.submission_02 = Submission.objects.create(
            type=self.submission_type, student=self.s2.student,
            text='I like apples')

        self.subscription = SubmissionSubscription.objects.create(
            owner=self.t, query_type=SubmissionSubscription.RANDOM)

    def test_subscription_gets_an_assignment(self):
        self.subscription.get_or_create_work_assignment()
        self.assertEqual(1, self.subscription.assignments.count())

    def test_first_work_assignment_was_created_unfinished(self):
        self.subscription.get_or_create_work_assignment()
        self.assertFalse(self.subscription.assignments.first().is_done)

    def test_subscription_raises_error_when_depleted(self):
        self.submission_01.delete()
        self.submission_02.delete()
        try:
            self.subscription.get_or_create_work_assignment()
        except SubscriptionEnded:
            self.assertFalse(False)
        else:
            self.assertTrue(False)

    def test_can_prefetch(self):
        self.subscription.get_or_create_work_assignment()
        self.subscription.get_or_create_work_assignment()
        self.assertEqual(2, self.subscription.assignments.count())

    def test_new_subscription_is_temporarily_unavailabe(self):
        validation = SubmissionSubscription.objects.create(
            owner=self.t, query_type=SubmissionSubscription.RANDOM,
            feedback_stage=SubmissionSubscription.FEEDBACK_VALIDATION)
        try:
            validation.get_or_create_work_assignment()
        except SubscriptionTemporarilyEnded:
            self.assertTrue(True)
        else:
            self.assertTrue(False)

    def test_delete_with_done_assignments_subscription_remains(self):
        first = self.subscription.get_or_create_work_assignment()
        self.subscription.get_or_create_work_assignment()
        self.assertEqual(2, self.subscription.assignments.count())

        first.is_done = True
        first.save()
        self.subscription.delete()
        self.assertEqual(1, self.subscription.assignments.count())

    def test_delete_without_done_assignments_subscription_is_deleted(self):
        self.subscription.delete()
        self.assertEqual(0, models.SubmissionSubscription.objects.count())

    def test_assignment_delete_of_done_not_permitted(self):
        first = self.subscription.get_or_create_work_assignment()
        first.is_done = True
        first.save()

        self.assertRaises(models.DeletionOfDoneAssignmentsNotPermitted,
                          first.delete)

    def test_assignment_delete_undone_permitted(self):
        first = self.subscription.get_or_create_work_assignment()
        first.delete()
        self.assertEqual(0, self.subscription.assignments.count())


class TestApiEndpoints(APITestCase):

    @classmethod
    def setUpTestData(cls):
        cls.data = make_test_data(data_dict={
            'submission_types': [
                {
                    'name': '01. Sort this or that',
                    'full_score': 35,
                    'description': 'Very complicated',
                    'solution': 'Trivial!'
                },
                {
                    'name': '02. Merge this or that or maybe even this',
                    'full_score': 35,
                    'description': 'Very complicated',
                    'solution': 'Trivial!'
                }
            ],
            'students': [
                {'username': 'student01'},
                {'username': 'student02'}
            ],
            'tutors': [
                {'username': 'tutor01'},
                {'username': 'tutor02'}
            ],
            'reviewers': [
                {'username': 'reviewer'}
            ],
            'submissions': [
                {
                    'text': 'function blabl\n'
                            '   on multi lines\n'
                            '       for blabla in bla:\n'
                            '           lorem ipsum und so\n',
                    'type': '01. Sort this or that',
                    'user': 'student01',
                    'feedback': {
                        'score': 5,
                        'is_final': True,
                        'feedback_lines': {
                            '1': [{
                                'text': 'This is very bad!',
                                'of_tutor': 'tutor01'
                            }],
                        }

                    }
                },
                {
                    'text': 'function blabl\n'
                            '       asasxasx\n'
                            '           lorem ipsum und so\n',
                    'type': '02. Merge this or that or maybe even this',
                    'user': 'student01'
                },
                {
                    'text': 'function blabl\n'
                            '   on multi lines\n'
                            '       asasxasx\n'
                            '           lorem ipsum und so\n',
                    'type': '01. Sort this or that',
                    'user': 'student02'
                },
                {
                    'text': 'function lorem ipsum etc\n',
                    'type': '02. Merge this or that or maybe even this',
                    'user': 'student02'
                },
            ]}
        )

    def setUp(self):
        self.client = APIClient()

    def test_ramaining_submissions_for_student(self):
        self.client.force_authenticate(user=self.data['reviewers'][0])

        student = self.data['students'][0]

        response = self.client.post(
            '/api/subscription/', {
                'query_type': 'student',
                'query_key': student.student.student_id,
                'stage': 'feedback-creation'
            })

        # This is expected since we wanted to assign all student submissions
        self.assertEqual(0, response.data['remaining'])
        self.assertEqual(0, response.data['available'])

    def test_remaining_submissions(self):
        self.client.force_authenticate(user=self.data['tutors'][0])

        response = self.client.post('/api/subscription/', {'query_type': 'random'})

        self.assertEqual(3, response.data['remaining'])

    def test_available_submissions_in_stage(self):
        self.client.force_authenticate(user=self.data['tutors'][0])

        response = self.client.post('/api/subscription/',
                                    {'query_type': 'random',
                                     'feedback_stage': 'feedback-validation'})

        self.assertEqual(0, response.data['available'])

    def test_can_create_a_subscription(self):
        self.client.force_authenticate(user=self.data['tutors'][0])

        response = self.client.post('/api/subscription/', {'query_type': 'random'})

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_create_subscription_and_get_one_assignment(self):
        self.client.force_authenticate(user=self.data['tutors'][0])

        response = self.client.post('/api/subscription/', {'query_type': 'random'})

        self.assertEqual('tutor01', response.data['owner'])

    def test_subscription_has_next_and_current_assignment(self):
        self.client.force_authenticate(user=self.data['tutors'][0])

        response_subscription_create = self.client.post(
            '/api/subscription/', {'query_type': 'random'})
        subscription_pk = response_subscription_create.data['pk']

        subscription_pk = response_subscription_create.data['pk']
        response_assignment = self.client.post(
            f'/api/assignment/', {
                'subscription': subscription_pk
            })

        assignment_pk = response_assignment.data['pk']
        response_subscription = self.client.get(
            f'/api/subscription/{subscription_pk}/')

        self.assertEqual(1, len(response_subscription.data['assignments']))
        self.assertEqual(response_assignment.data['pk'],
                         response_subscription.data['assignments'][0]['pk'])

        subscription_pk = response_subscription.data['pk']
        response_next = self.client.post(
            f'/api/assignment/', {
                'subscription': subscription_pk
            })
        response_detail_subs = \
            self.client.get(f'/api/subscription/{subscription_pk}/')

        self.assertEqual(2, len(response_detail_subs.data['assignments']))
        self.assertNotEqual(assignment_pk, response_next.data['pk'])

    def test_subscription_can_assign_to_student(self):
        self.client.force_authenticate(user=self.data['reviewers'][0])

        student = self.data['students'][0]

        response = self.client.post(
            '/api/subscription/', {
                'query_type': 'student',
                'query_key': student.student.student_id,
                'stage': 'feedback-creation'
            })

        assignments = response.data['assignments']
        self.assertEqual(1, len(assignments))

    def test_two_tutors_cant_have_assignments_for_same_submission(self):
        self.client.force_authenticate(user=self.data['tutors'][0])

        subscription = self.client.post('/api/subscription/',
                                        {'query_type': 'random'}).data

        assignment_fst_tutor = self.client.post(
            '/api/assignment/', {
                'subscription': subscription['pk']
            }
        ).data

        self.client.force_authenticate(user=self.data['tutors'][1])

        subscription = self.client.post('/api/subscription/',
                                        {'query_type': 'random'}).data

        assignment_snd_tutor = self.client.post(
            '/api/assignment/', {
                'subscription': subscription['pk']
            }
        ).data

        self.assertNotEqual(assignment_fst_tutor['submission']['pk'],
                            assignment_snd_tutor['submission']['pk'])

    def test_reviewer_can_get_active_assignments(self):
        self.client.force_authenticate(user=self.data['tutors'][0])
        subscription = self.client.post('/api/subscription/',
                                        {'query_type': 'random'}).data

        assignment = self.client.post(
            '/api/assignment/', {
                'subscription': subscription['pk']
            }
        ).data

        # tutors shouldn't have access
        res = self.client.get(
            '/api/assignment/active/'
        )
        self.assertEqual(status.HTTP_403_FORBIDDEN, res.status_code)

        self.client.force_authenticate(user=self.data['reviewers'][0])
        active_assignments = self.client.get(
            '/api/assignment/active/'
        ).data
        print(assignment)
        self.assertIn(assignment['pk'], [assignment['pk'] for assignment in active_assignments])

    def test_reviewer_can_delete_active_assignments(self):
        self.client.force_authenticate(user=self.data['tutors'][0])
        subscription = self.client.post('/api/subscription/',
                                        {'query_type': 'random'}).data

        assignment = self.client.post(
            '/api/assignment/', {
                'subscription': subscription['pk']
            }
        ).data

        # tutors shouldn't have access
        res = self.client.delete(
            '/api/assignment/active/'
        )
        self.assertEqual(status.HTTP_403_FORBIDDEN, res.status_code)

        self.client.force_authenticate(user=self.data['reviewers'][0])
        res = self.client.delete(
            '/api/assignment/active/'
        )
        self.assertEqual(status.HTTP_204_NO_CONTENT, res.status_code)
        self.assertNotIn(
            assignment['pk'],
            [assignment.pk for assignment
             in TutorSubmissionAssignment.objects.filter(is_done=False)]
        )

    def test_all_stages_of_the_subscription(self):
        self.client.force_authenticate(user=self.data['tutors'][0])

        # The tutor corrects something
        response = self.client.post(
            '/api/subscription/', {
                'query_type': 'random',
                'feedback_stage': 'feedback-creation'
            })

        self.assertEqual(status.HTTP_201_CREATED, response.status_code)

        subscription_pk = response.data['pk']
        response = self.client.post(
            f'/api/assignment/', {
                'subscription': subscription_pk
            })
        self.assertEqual(status.HTTP_201_CREATED, response.status_code)
        response = self.client.post(
            f'/api/feedback/', {
                "score": 23,
                "of_submission": response.data['submission']['pk'],
                "feedback_lines": {
                    1: {"text": "< some string >"},
                    2: {"text": "< some string >"}
                }
            }
        )
        self.assertEqual(status.HTTP_201_CREATED, response.status_code)

        # some other tutor reviews it
        self.client.force_authenticate(user=self.data['tutors'][1])

        response = self.client.post(
            '/api/subscription/', {
                'query_type': 'random',
                'feedback_stage': 'feedback-validation'
            })

        self.assertEqual(status.HTTP_201_CREATED, response.status_code)

        subscription_pk = response.data['pk']
        response = self.client.post(
            '/api/assignment/', {
                'subscription': subscription_pk
            })

        self.assertEqual(status.HTTP_201_CREATED, response.status_code)
        submission_id_in_database = models.Feedback.objects.filter(
            is_final=False).first().of_submission.submission_id
        submission_id_in_response = response.data['submission']['pk']

        self.assertEqual(
            str(submission_id_in_database),
            submission_id_in_response)

        assignment = models.TutorSubmissionAssignment.objects.get(
            pk=response.data['pk'])
        self.assertFalse(assignment.is_done)
        response = self.client.patch(
            '/api/feedback/%s/' % submission_id_in_response, {
                "score": 20,
                "is_final": True,
                "feedback_lines": {
                    2: {"text": "< some addition by second tutor>"},
                }
            }
        )

        assignment.refresh_from_db()
        meta = assignment.submission.meta
        self.assertEqual(status.HTTP_200_OK, response.status_code)
        self.assertEqual(2, len(response.data['feedback_lines'][2]))
        self.assertTrue(assignment.is_done)
        self.assertIn(self.data['tutors'][0], meta.feedback_authors.all())
        self.assertIn(self.data['tutors'][1], meta.feedback_authors.all())
