import Vue from 'vue'
import * as hljs from 'highlight.js'
import * as api from '@/api'
import { Feedback, FeedbackComment, SubmissionNoType } from '@/models'
import { RootState } from '@/store/store'
import { getStoreBuilder, BareActionContext } from 'vuex-typex'

export interface SubmissionNotesState {
  submission: SubmissionNoType
  ui: {
      showEditorOnLine: {[lineNo: number]: boolean}
      selectedCommentOnLine: {[lineNo: number]: FeedbackComment}
      showFeedback: boolean
  },
  hasOrigFeedback: boolean
  origFeedback: Feedback
  updatedFeedback: Feedback
  commentsMarkedForDeletion: {[pk: string]: FeedbackComment}
}

function initialState (): SubmissionNotesState {
  return {
    submission: {
      text: '',
      pk: '',
      type: '',
      tests: []
    },
    ui: {
      showEditorOnLine: {},
      selectedCommentOnLine: {},
      showFeedback: true
    },
    hasOrigFeedback: false,
    origFeedback: {
      pk: 0,
      score: undefined,
      isFinal: false,
      feedbackLines: {}
    },
    updatedFeedback: {
      pk: 0,
      score: undefined,
      feedbackLines: {}
    },
    commentsMarkedForDeletion: {}
  }
}

const mb = getStoreBuilder<RootState>().module('SubmissionNotes', initialState())

const stateGetter = mb.state()

const submissionTypeGetter = mb.read(function submissionType (state, getters, rootState) {
  return rootState.submissionTypes[state.submission.type]
})
// highlight the submission the reduce the string
// submission.text into an object where the keys are the
// line indexes starting at one and the values the corresponding submission line
// this makes iterating over the submission much more pleasant
const submissionGetter = mb.read(function submission (state, getters) {
  const language = getters.submissionType
    ? getters.submissionType.programmingLanguage
    : 'c'
  const highlighted = hljs.highlight(language, state.submission.text || '', true).value
  return highlighted.split('\n').reduce((acc: {[k: number]: string}, cur, index) => {
    acc[index + 1] = cur
    return acc
  }, {})
})
const scoreGetter = mb.read(function score (state) {
  return state.updatedFeedback.score !== undefined ? state.updatedFeedback.score : state.origFeedback.score
})
const workInProgressGetter = mb.read(function workInProgress (state) {
  const openEditor = Object.values(state.ui.showEditorOnLine).reduce((acc, curr) => acc || curr, false)
  const feedbackWritten = Object.entries(state.updatedFeedback.feedbackLines || {}).length > 0
  return openEditor || feedbackWritten
})
const isFeedbackCreationGetter = mb.read(function isFeedbackCreation (state) {
  return !state.origFeedback['feedbackStageForUser'] ||
    state.origFeedback['feedbackStageForUser'] === 'feedback-creation'
})

function SET_SUBMISSION (state: SubmissionNotesState, submission: SubmissionNoType) {
  state.submission = submission
}
function SET_ORIG_FEEDBACK (state: SubmissionNotesState, feedback: Feedback) {
  if (feedback) {
    state.origFeedback = feedback
    state.hasOrigFeedback = true
  }
}
function SET_SHOW_FEEDBACK (state: SubmissionNotesState, val: boolean) {
  state.ui.showFeedback = val
}
function UPDATE_FEEDBACK_LINE (state: SubmissionNotesState, feedback: {lineNo: number, comment: Partial<FeedbackComment>}) {
  // explicit .toString() on lineNo is necessary because of Vue.set typings
  if (state.updatedFeedback.feedbackLines) {
    Vue.set(state.updatedFeedback.feedbackLines, feedback.lineNo.toString(), feedback.comment)
  }
}
function UPDATE_FEEDBACK_SCORE (state: SubmissionNotesState, score: number) {
  state.updatedFeedback.score = score
}
function DELETE_FEEDBACK_LINE (state: SubmissionNotesState, lineNo: number) {
  if (state.updatedFeedback.feedbackLines) {
    // Vue dosn't like objects that are indexed with numbers...
    Vue.delete(state.updatedFeedback.feedbackLines, <string><any>lineNo)
  }
}
function TOGGLE_EDITOR_ON_LINE (state: SubmissionNotesState, { lineNo, comment }: {lineNo: number, comment: FeedbackComment}) {
  Vue.set(state.ui.selectedCommentOnLine, <string><any>lineNo, comment)
  Vue.set(state.ui.showEditorOnLine, <string><any> lineNo, !state.ui.showEditorOnLine[lineNo])
}
function MARK_COMMENT_FOR_DELETION (state: SubmissionNotesState, comment: FeedbackComment) {
  Vue.set(state.commentsMarkedForDeletion, comment.pk, comment)
}
function UN_MARK_COMMENT_FOR_DELETION (state: SubmissionNotesState, comment: FeedbackComment) {
  Vue.delete(state.commentsMarkedForDeletion, comment.pk)
}
function RESET_UPDATED_FEEDBACK (state: SubmissionNotesState) {
  state.updatedFeedback = initialState().updatedFeedback
}
function RESET_STATE (state: SubmissionNotesState) {
  Object.assign(state, initialState())
}

async function deleteComments ({ state }: BareActionContext<SubmissionNotesState, RootState>) {
  return Promise.all(
    Object.values(state.commentsMarkedForDeletion).map(comment => {
      return api.deleteComment(comment)
    })
  )
}
async function submitFeedback ({ state }: BareActionContext<SubmissionNotesState, RootState>, { isFinal = false }) {
  let feedback: Partial<Feedback> = {
    isFinal: isFinal,
    ofSubmission: state.submission.pk
  }
  if (state.origFeedback.score === undefined && state.updatedFeedback.score === undefined) {
    throw new Error('You need to give a score.')
  } else if (state.updatedFeedback.score !== undefined) {
    feedback.score = state.updatedFeedback.score
  } else {
    feedback.score = state.origFeedback.score
  }

  if (Object.keys(state.updatedFeedback.feedbackLines || {}).length > 0) {
    feedback.feedbackLines = state.updatedFeedback.feedbackLines
  } else if (feedback.score! < SubmissionNotes.submissionType.fullScore!) {
    throw new Error('You need to add or change a comment when setting a non full score.')
  }
  // delete those comments that have been marked for deletion
  await SubmissionNotes.deleteComments()
  if (!state.hasOrigFeedback) {
    return api.submitFeedbackForAssignment({ feedback })
  } else {
    feedback.pk = state.origFeedback.pk
    return api.submitUpdatedFeedback(<{feedback: Feedback}> { feedback })
  }
}

export const SubmissionNotes = {
  get state () { return stateGetter() },
  get submissionType () { return submissionTypeGetter() },
  get submission () { return submissionGetter() },
  get score () { return scoreGetter() },
  get workInProgress () { return workInProgressGetter() },
  get isFeedbackCreation () { return isFeedbackCreationGetter() },

  SET_SUBMISSION: mb.commit(SET_SUBMISSION),
  SET_ORIG_FEEDBACK: mb.commit(SET_ORIG_FEEDBACK),
  SET_SHOW_FEEDBACK: mb.commit(SET_SHOW_FEEDBACK),
  UPDATE_FEEDBACK_LINE: mb.commit(UPDATE_FEEDBACK_LINE),
  UPDATE_FEEDBACK_SCORE: mb.commit(UPDATE_FEEDBACK_SCORE),
  DELETE_FEEDBACK_LINE: mb.commit(DELETE_FEEDBACK_LINE),
  TOGGLE_EDITOR_ON_LINE: mb.commit(TOGGLE_EDITOR_ON_LINE),
  MARK_COMMENT_FOR_DELETION: mb.commit(MARK_COMMENT_FOR_DELETION),
  UN_MARK_COMMENT_FOR_DELETION: mb.commit(UN_MARK_COMMENT_FOR_DELETION),
  RESET_UPDATED_FEEDBACK: mb.commit(RESET_UPDATED_FEEDBACK),
  RESET_STATE: mb.commit(RESET_STATE),

  deleteComments: mb.dispatch(deleteComments),
  submitFeedback: mb.dispatch(submitFeedback)
}
