from django.test import LiveServerTestCase
from selenium import webdriver
from selenium.webdriver.support.ui import WebDriverWait

from core import models
from core.models import UserAccount
from functional_tests.util import (login, create_browser, subscriptions_loaded_cond,
                                   extract_hrefs_hashes, reset_browser_after_test)
from util import factory_boys as fact


# This is a little hack to have Super test class which implements common behaviour
# and tests but is not executed. In order to have the testrunner ignore the
# FrontPageTestsTutorReviewer class we need to define it within a class which does not inherit from
# unittest
class UntestedParent:
    class FrontPageTestsTutorReviewer(LiveServerTestCase):
        browser: webdriver.Firefox = None
        username = None
        password = None
        role = None

        @classmethod
        def setUpClass(cls):
            super().setUpClass()
            cls.browser = create_browser()

        @classmethod
        def tearDownClass(cls):
            super().tearDownClass()
            cls.browser.quit()

        def setUp(self):
            fact.SubmissionFactory.create_batch(4)

        def tearDown(self):
            reset_browser_after_test(self.browser, self.live_server_url)

        def _login(self):
            login(self.browser, self.live_server_url, self.username, self.password)

        def test_statistics_are_shown(self):
            self._login()
            statistics = self.browser.find_element_by_id('correction-statistics')
            title = statistics.find_element_by_class_name('title')
            self.assertEqual('Statistics', title.text)

        def test_available_tasks_are_shown(self):
            self._login()
            tasks = self.browser.find_element_by_name('subscription-list')
            title = tasks.find_element_by_class_name('v-toolbar__title')
            self.assertEqual('Tasks', title.text)
            WebDriverWait(self.browser, 6).until(subscriptions_loaded_cond(tasks))
            subscription_links = extract_hrefs_hashes(
                tasks.find_elements_by_tag_name('a')
            )
            subscriptions = models.SubmissionSubscription.objects.filter(
                owner__username=self.username,
                deactivated=False,
                feedback_stage=models.SubmissionSubscription.FEEDBACK_CREATION
            ).exclude(query_type=models.SubmissionSubscription.RANDOM)
            for sub in subscriptions:
                self.assertIn(f'/subscription/{sub.pk}', subscription_links)


class FrontPageTestsTutor(UntestedParent.FrontPageTestsTutorReviewer):
    def setUp(self):
        super().setUp()
        self.username = 'tutor'
        self.password = 'p'
        self.role = UserAccount.TUTOR
        fact.UserAccountFactory(
            username=self.username,
            password=self.password
        )

    def test_side_bar_contains_correct_items(self):
        self._login()
        drawer = self.browser.find_element_by_class_name('v-navigation-drawer')
        links = extract_hrefs_hashes(drawer.find_elements_by_tag_name('a'))
        print(links)
        self.assertTrue(all(link in links for link in ['/home', '/feedback']))
        task_title = drawer.find_element_by_class_name('v-toolbar__title')
        self.assertEqual('Tasks', task_title.text)
        footer = drawer.find_element_by_class_name('sidebar-footer')
        feedback_link = footer.find_element_by_css_selector('a.feedback-link')
        self.assertEqual('Give us Feedback!', feedback_link.text)
        self.assertEqual('https://gitlab.gwdg.de/j.michal/grady/issues',
                         feedback_link.get_attribute('href'))


class FrontPageTestsReviewer(UntestedParent.FrontPageTestsTutorReviewer):
    def setUp(self):
        super().setUp()
        self.username = 'reviewer'
        self.password = 'p'
        self.role = UserAccount.REVIEWER
        fact.UserAccountFactory(
            username=self.username,
            password=self.password,
            role=self.role
        )

    def test_side_bar_contains_correct_items(self):
        self._login()
        drawer = self.browser.find_element_by_class_name('v-navigation-drawer')
        links = extract_hrefs_hashes(drawer.find_elements_by_tag_name('a'))
        self.assertTrue(all(link in links for link in
                            ['/home', '/feedback', '/student-overview', '/tutor-overview']))
        task_title = drawer.find_element_by_class_name('v-toolbar__title')
        self.assertEqual('Tasks', task_title.text)
        footer = drawer.find_element_by_class_name('sidebar-footer')
        feedback_link = footer.find_element_by_css_selector('a.feedback-link')
        self.assertEqual('Give us Feedback!', feedback_link.text)
        self.assertEqual('https://gitlab.gwdg.de/j.michal/grady/issues',
                         feedback_link.get_attribute('href'))
